import 'dart:async';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_mini_program/engine/nui_miniprogram_js_engine.dart';
import 'package:nui_mini_program/model/nui_miniprogram_module.dart';
import 'package:nui_mini_program/util/nui_miniprogram_action_util.dart';
import 'package:nui_mini_program/util/nui_miniprogram_file_manager.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_enum.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_message.dart';
import 'package:nui_miniprogram_core/nui_miniprogram_engine.dart';
import 'package:webview_flutter_plus/webview_flutter_plus.dart';

class NUIMiniProgramView extends StatefulWidget {
  NUIMiniProgramView({@required this.module, @required this.controller, this.showStatusBar: true, this.placeHolder, this.loadUntilInitialized: false});
  final NUIMiniProgramModule module;
  final NUIMiniProgramController controller;
  final bool showStatusBar;
  final WidgetBuilder placeHolder;
  final bool loadUntilInitialized;
  @override
  _NUIMiniProgramViewState createState() => _NUIMiniProgramViewState();
}

class _NUIMiniProgramViewState extends State<NUIMiniProgramView> {
  static const MODULE = "NUIMiniProgramBuilder";

  NUIMiniProgramController _controller;
  NUIMiniProgramFileManager _fileManager;
  NUIMiniProgramJSEngine _jsEngine;
  Future<String> _moduleIndexHtml;
  String _currentIndexFile;

  bool _firstLoaded = false;

  @override
  void initState(){
    super.initState();
    _controller = widget.controller;
    _fileManager = NUIMiniProgramFileManager(widget.module);
    _jsEngine = NUIMiniProgramJSEngine.from(widget.module);
    _moduleIndexHtml = _fileManager.initializeMiniProgram();

    _controller._jsEngine = _jsEngine;
    _initChannels();
  }

  void _initChannels(){
    final channels = <NUIMiniProgramChannel>[];
    channels.addAll(_controller.channels);
    channels.add(NUIMiniProgramChannel(NUIMiniProgramMessage.DEFAULT_CHANNEL, listener: (message, msg) async{
      return NUIMiniProgramActionUtil().processAction(context, message, msg, initializeListener: (){
        if(widget.loadUntilInitialized == true && _firstLoaded != true) {
          setState(() {
            _firstLoaded = true;
          });
        }
      });
    }));
    _jsEngine.setChannels(channels);
  }

  @override
  void dispose(){
    super.dispose();
    if(!isNullOrEmpty(_currentIndexFile)) {
      _fileManager.disableDynamicAssets(_currentIndexFile);
    }
    _controller._jsEngine = null;
  }

  Future<bool> _overrideBackNavigation() async{
    final controller = await _controller.webController;
    final canGoBack = await controller.webViewController.canGoBack();
    if(canGoBack){
      controller.webViewController.goBack();
      return false;
    }
    else{
      final maxDelay = NUIAsync.delayForResult(Duration(milliseconds: 100), () async => true);
      final completer = Completer<bool>();
      _controller.canMiniProgramPop().then((value) {
        completer.complete(value);
      });
      await maxDelay;
      if(completer.isCompleted){
        final canMiniPop = await completer.future;
        if(canMiniPop){
          return false;
        }
        return true;
      }
      return true;
    }
  }

  Widget _loadingPlaceholder(BuildContext context){
    if(widget.placeHolder != null){
      return widget.placeHolder(context);
    }
    if(widget.module.getIconUrl() != null) {
      return Center(child: loadImage(widget.module.getIconUrl(), width: 100, height: 100, placeHolder: SizedBox(width: 1, height: 1)));
    }
    else{
      return Center(child: CircularProgressIndicator());
    }
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      body: WillPopScope(
        onWillPop: () async{
          return _overrideBackNavigation();
        },
        child: Container(
          width: double.infinity,
          height: double.infinity,
          decoration: NUIBoxDecoration(
            gradient: NUILinearGradient(
              colors: widget.module.getBrandColors() ?? [NUIColors.NUIWhite, NUIColors.NUISkeletalGray],
              orientation: GradientOrientation.TR_BL
            ),
          ),
          child: Column(
            mainAxisSize: MainAxisSize.max,
            children: [
              widget.showStatusBar == true ? NUIStatusBar() : null,
              Expanded(
                child: Stack(
                  children: [
                    AnimatedOpacity(
                      duration: Duration(milliseconds: 250),
                      opacity: _firstLoaded ? 0 : 1,
                      child: Container(
                        width: double.infinity,
                        height: double.infinity,
                        child: _loadingPlaceholder(context),
                      ),
                    ),
                    NUIFutureWidget(
                        futureValue: _moduleIndexHtml,
                        callback: (state, html, message){
                          if(state == FutureState.SUCCESS && !isNullOrEmpty(html)){
                            _currentIndexFile = html;
                            return Builder(builder: (BuildContext context) {
                              logNUI(MODULE, "Rendering html from : $html");
                              _fileManager.enableDynamicAssets(html);
                              return Container(
                                width: double.infinity,
                                height: double.infinity,
                                child: AnimatedOpacity(
                                  duration: Duration(milliseconds: 250),
                                  opacity: _firstLoaded ? 1 : 0,
                                  child: WebViewPlus(
                                    initialUrl: html,
                                    javascriptMode: JavascriptMode.unrestricted,
                                    javascriptChannels: [
                                      _jsEngine.getJavascriptChannel("NUIMiniProgram"),
                                    ].toSet(),
                                    onWebViewCreated: (WebViewPlusController webViewController) {
                                      _jsEngine.setWebController(webViewController);
                                      _controller._setWebController(webViewController);
                                    },
                                    onPageStarted: (url){
                                      logNUI(MODULE, "On page started at url: $url");
                                    },
                                    onPageFinished: (url){
                                      if(widget.loadUntilInitialized != true && _firstLoaded != true) {
                                        setState(() {
                                          _firstLoaded = true;
                                        });
                                      }
                                      // logNUI(MODULE, "On page finished at url: $url");
                                    },
                                    // gestureRecognizers: Set()
                                    //   ..add(
                                    //       Factory<LongPressGestureRecognizer>(() => LongPressGestureRecognizer()..onLongPress = (){
                                    //         logNUI("NUIMiniProgramView", "On long press detected");
                                    //       })
                                    //   )
                                  ),
                                ),
                              );
                            });
                          }
                          else{
                            return SizedBox.shrink();
                          }
                        }
                    ),
                  ],
                ),
              ),
            ],
          ),
        ),
      ),
    );
  }
}


class NUIMiniProgramController{
  static const MODULE = "NUIMiniProgramController";
  NUIMiniProgramController({List<NUIMiniProgramChannel> channels}){
    _channels = channels;
  }

  NUIMiniProgramJSEngineInterface _jsEngine;
  Completer<WebViewPlusController> _webController = Completer();
  List<NUIMiniProgramChannel> _channels;

  Future<WebViewPlusController> get webController => _webController.future;

  Future<NUIMiniProgramMessage> postMessage(NUIMiniProgramMessage message) async{
    try {
      await _webController.future;
      return _postMessageToMiniProgram(message);
    }catch(e){
      logNUI(MODULE, "Failed to encode message with error : $e");
      return null;
    }
  }
  
  Future<bool> canMiniProgramPop() async{
    final result = await _jsEngine?.sendAndAwaitCrossEvent(NUIMiniProgramMessage(channel: NUIMiniProgramWebEventType.CAN_POP.value, data: NUIMiniProgramWebEventType.CAN_POP.value));
    logNUI("NUIMiniProgram", "Can pop result : ${result.data}");
    if(result != null){
      return match(result.data, "true");
    }
    return false;
  }

  Future<NUIMiniProgramMessage> _postMessageToMiniProgram(NUIMiniProgramMessage message) async{
    final result = await _jsEngine?.sendAndAwaitCrossEvent(message);
    return result;
  }

  void _setWebController(WebViewPlusController controller){
    _webController.complete(controller);
  }

  List<NUIMiniProgramChannel> get channels => _channels ?? [];
}
